use super::{BrokerError, BrokerResult};
use crate::{Message, Sender};
use log::trace;
use std::{
    any::{Any, TypeId},
    borrow::Cow,
    collections::HashMap,
    marker::PhantomData,
};
use tokio::sync::{broadcast, mpsc, oneshot};

pub trait MessageType {
    fn message_type_id(&self) -> TypeId;

    fn message_type_name(&self) -> &'static str;

    fn create_broadcast_sender(&self) -> ErasedSender;

    fn create_mpsc_sender(&self) -> ErasedSender;
}

pub struct BasicMessageType<T> {
    _phantom: PhantomData<T>,
}

impl<T> Default for BasicMessageType<T> {
    fn default() -> Self {
        BasicMessageType {
            _phantom: PhantomData,
        }
    }
}

impl<T: Message> MessageType for BasicMessageType<T> {
    fn message_type_id(&self) -> TypeId {
        TypeId::of::<T>()
    }

    fn message_type_name(&self) -> &'static str {
        std::any::type_name::<T>()
    }

    fn create_broadcast_sender(&self) -> ErasedSender {
        trace!(
            "Creating broadcast sender for {} ({:?})",
            std::any::type_name::<T>(),
            MessageType::type_id(self)
        );
        // TODO: configurable queue size (per message?)
        let (sender, _) = broadcast::channel::<T>(8);
        let sender: ErasedSender = Box::new(sender);
        sender
    }

    fn create_mpsc_sender(&self) -> ErasedSender {
        trace!(
            "Creating mpsc sender for {} ({:?})",
            std::any::type_name::<T>(),
            MessageType::type_id(self)
        );
        Box::new(MpscSender::<T>::new())
    }
}

pub type ErasedSender = Box<dyn Any + Send + Sync>;

pub struct MpscSender<T> {
    sender: mpsc::Sender<T>,
    waiting: Option<MpscWaiting<T>>,
}

struct MpscWaiting<T> {
    receiver: mpsc::Receiver<T>,
    waiting: Vec<oneshot::Sender<BrokerResult<Sender<T>>>>,
}

impl<T: Message> MpscSender<T> {
    fn new() -> Self {
        // TODO: configurable queue size (per message?)
        let (sender, receiver) = mpsc::channel(8);
        MpscSender {
            sender,
            waiting: Some(MpscWaiting {
                receiver,
                waiting: Vec::default(),
            }),
        }
    }

    pub fn clone_sender_or_wait(&mut self, wait: oneshot::Sender<BrokerResult<Sender<T>>>) {
        if let Some(waiting) = &mut self.waiting {
            waiting.waiting.push(wait);
        } else {
            let _ = wait.send(Ok(self.clone_sender()));
        }
    }

    pub fn clone_sender(&self) -> Sender<T> {
        Sender::new(self.sender.clone())
    }

    pub fn claim_receiver(&mut self) -> Option<mpsc::Receiver<T>> {
        self.waiting.take().map(|w| {
            for response_tx in w.waiting.into_iter() {
                let _ = response_tx.send(Ok(self.clone_sender()));
            }
            w.receiver
        })
    }
}

#[derive(Debug)]
pub struct TopicEntry {
    message_type_id: TypeId,
    message_type_name: String,
    sender: ErasedSender,
}

#[derive(Debug)]
pub struct Registry {
    topics: HashMap<Cow<'static, str>, TopicEntry>,
}

impl Default for Registry {
    fn default() -> Self {
        Registry {
            topics: HashMap::new(),
        }
    }
}

impl Registry {
    fn get_erased_sender_for_type<'a, 'b>(
        &'a mut self,
        message_type: &'b dyn MessageType,
        create_sender: fn(&'b dyn MessageType) -> ErasedSender,
    ) -> &'a mut ErasedSender {
        let type_id = message_type.message_type_id();
        let type_name = message_type.message_type_name();
        let topic_entry = self.topics.entry(type_name.into());
        let topic_entry = topic_entry.or_insert_with(|| TopicEntry {
            message_type_id: type_id,
            message_type_name: type_name.to_string(),
            sender: create_sender(message_type),
        });
        &mut topic_entry.sender
    }

    fn get_sender_for_type<'a, 'b, T: 'static>(
        &'a mut self,
        message_type: &'b dyn MessageType,
        create_sender: fn(&'b dyn MessageType) -> ErasedSender,
    ) -> BrokerResult<&'a mut T> {
        let erased = self.get_erased_sender_for_type(message_type, create_sender);
        (**erased)
            .downcast_mut::<T>()
            .ok_or(BrokerError::MismatchedType)
    }

    pub fn get_broadcast_sender<T: Message>(
        &mut self,
        message_type: &dyn MessageType,
    ) -> BrokerResult<&mut broadcast::Sender<T>> {
        self.get_sender_for_type(message_type, MessageType::create_broadcast_sender)
    }

    pub fn get_mpsc_sender<T: Message>(
        &mut self,
        message_type: &dyn MessageType,
    ) -> BrokerResult<&mut MpscSender<T>> {
        self.get_sender_for_type(message_type, MessageType::create_mpsc_sender)
    }
}
