use std::marker::PhantomData;
use tokio::sync::{mpsc, oneshot};

use super::{
    registry::{BasicMessageType, Registry},
    BrokerError, BrokerResult,
};
use crate::{Message, Publication, Receiver, Sender, Subscription};

pub trait BrokerRequestInternal: Send {
    fn run_request(self: Box<Self>, registry: &mut Registry);
}

pub trait MakeBrokerRequest: BrokerRequestInternal {
    type Response;

    fn create(response_tx: oneshot::Sender<Self::Response>) -> Box<Self>;

    fn new() -> (Box<Self>, oneshot::Receiver<Self::Response>) {
        let (response_tx, response_rx) = oneshot::channel();
        (Self::create(response_tx), response_rx)
    }
}

pub trait RequestHandler: Send {
    type Response: Send;

    fn handle(self, registry: &mut Registry) -> Self::Response;
}

pub struct BrokerRequest<H: RequestHandler> {
    handler: H,
    response_tx: oneshot::Sender<H::Response>,
}

impl<H: RequestHandler> BrokerRequestInternal for BrokerRequest<H> {
    fn run_request(self: Box<Self>, registry: &mut Registry) {
        let response = self.handler.handle(registry);
        let _ = self.response_tx.send(response);
    }
}

impl<H: RequestHandler + Default> MakeBrokerRequest for BrokerRequest<H> {
    type Response = H::Response;

    fn create(response_tx: oneshot::Sender<H::Response>) -> Box<Self> {
        Box::new(Self {
            handler: H::default(),
            response_tx,
        })
    }
}

pub type BrokerRequestBox = Box<dyn BrokerRequestInternal>;
pub type BrokerRequestSender = mpsc::Sender<BrokerRequestBox>;

pub type SubscribeRequest<T> = BrokerRequest<Subscribe<T>>;
pub struct Subscribe<T: Message> {
    msg_type: BasicMessageType<T>,
}

impl<T: Message> Default for Subscribe<T> {
    fn default() -> Self {
        Self {
            msg_type: BasicMessageType::<T>::default(),
        }
    }
}

impl<T: Message> RequestHandler for Subscribe<T> {
    type Response = BrokerResult<Subscription<T>>;

    fn handle(self, registry: &mut Registry) -> Self::Response {
        let sender = registry.get_broadcast_sender::<T>(&self.msg_type);
        sender.map(|sender| {
            let receiver = sender.subscribe();
            Subscription::new(receiver)
        })
    }
}

pub type AdvertiseRequest<T> = BrokerRequest<Advertise<T>>;
pub struct Advertise<T: Message> {
    msg_type: BasicMessageType<T>,
}

impl<T: Message> Default for Advertise<T> {
    fn default() -> Self {
        Self {
            msg_type: BasicMessageType::<T>::default(),
        }
    }
}

impl<T: Message> RequestHandler for Advertise<T> {
    type Response = BrokerResult<Publication<T>>;

    fn handle(self, registry: &mut Registry) -> Self::Response {
        let sender = registry.get_broadcast_sender::<T>(&self.msg_type);
        sender.map(|sender| Publication::new(sender.clone()))
    }
}

pub type ClaimReceiverRequest<T> = BrokerRequest<ClaimReceiver<T>>;

pub enum ClaimReceiverError {
    AlreadyClaimed,
    Broker(BrokerError),
}

pub struct ClaimReceiver<T: Message> {
    msg_type: BasicMessageType<T>,
}

impl<T: Message> Default for ClaimReceiver<T> {
    fn default() -> Self {
        Self {
            msg_type: BasicMessageType::<T>::default(),
        }
    }
}

impl<T: Message> RequestHandler for ClaimReceiver<T> {
    type Response = Result<Receiver<T>, ClaimReceiverError>;

    fn handle(self, registry: &mut Registry) -> Self::Response {
        registry
            .get_mpsc_sender::<T>(&self.msg_type)
            .map_err(ClaimReceiverError::Broker)
            .and_then(|sender| {
                sender
                    .claim_receiver()
                    .ok_or(ClaimReceiverError::AlreadyClaimed)
            })
            .map(Receiver::new)
    }
}

pub struct SenderRequest<T, W> {
    msg_type: BasicMessageType<T>,
    response_tx: oneshot::Sender<BrokerResult<Sender<T>>>,
    _wait: PhantomData<W>,
}

pub trait GetSenderWait: Send {
    const WAIT: bool;
}

pub struct NoWait;
impl GetSenderWait for NoWait {
    const WAIT: bool = false;
}
pub type SenderRequestNoWait<T> = SenderRequest<T, NoWait>;

pub struct Wait;
impl GetSenderWait for Wait {
    const WAIT: bool = true;
}
pub type SenderRequestWait<T> = SenderRequest<T, Wait>;

impl<T: Message, W: GetSenderWait> MakeBrokerRequest for SenderRequest<T, W> {
    type Response = BrokerResult<Sender<T>>;

    fn create(response_tx: oneshot::Sender<Self::Response>) -> Box<Self> {
        Box::new(Self {
            msg_type: BasicMessageType::default(),
            response_tx,
            _wait: PhantomData,
        })
    }
}

impl<T: Message, W: GetSenderWait> BrokerRequestInternal for SenderRequest<T, W> {
    fn run_request(self: Box<Self>, registry: &mut Registry) {
        let response = match registry.get_mpsc_sender::<T>(&self.msg_type) {
            Ok(sender) => Ok(if W::WAIT {
                sender.clone_sender_or_wait(self.response_tx);
                return;
            } else {
                sender.clone_sender()
            }),
            Err(err) => Err(err),
        };
        let _ = self.response_tx.send(response);
    }
}
