#![warn(clippy::all)]

mod broker_task;
mod message;
pub mod mpsc;
pub mod pub_sub;

use broker_task::{BrokerError, BrokerRequestSender, BrokerResult, BrokerTask};
pub use message::Message;
pub use mpsc::*;
pub use pub_sub::*;

use futures::executor::block_on;

#[derive(Clone, Debug, PartialEq)]
#[non_exhaustive]
pub enum OrsbError {
    BrokerClosed,
    NoResponse,
    MismatchedType,
}

impl From<BrokerError> for OrsbError {
    fn from(err: BrokerError) -> Self {
        match err {
            BrokerError::MismatchedType => OrsbError::MismatchedType,
        }
    }
}

fn map_broker_response<T>(result: Result<BrokerResult<T>, OrsbError>) -> Result<T, OrsbError> {
    match result {
        Ok(Ok(value)) => Ok(value),
        Ok(Err(err)) => Err(err.into()),
        Err(_) => Err(OrsbError::NoResponse),
    }
}

#[derive(Clone, Debug, PartialEq)]
#[non_exhaustive]
pub enum ClaimReceiverError {
    AlreadyClaimed,
    Other(OrsbError),
}

impl From<broker_task::ClaimReceiverError> for ClaimReceiverError {
    fn from(err: broker_task::ClaimReceiverError) -> Self {
        match err {
            broker_task::ClaimReceiverError::AlreadyClaimed => ClaimReceiverError::AlreadyClaimed,
            broker_task::ClaimReceiverError::Broker(err) => ClaimReceiverError::Other(err.into()),
        }
    }
}

#[derive(Debug, Clone)]
pub struct Orsb {
    request_tx: BrokerRequestSender,
}

impl Orsb {
    pub fn start_new() -> Self {
        Self::start_new2().0
    }

    pub(crate) fn start_new2() -> (Self, tokio::task::JoinHandle<()>) {
        let (request_tx, join_handle) = BrokerTask::start();

        let orsb = Orsb { request_tx };

        (orsb, join_handle)
    }

    async fn make_request<R: broker_task::MakeBrokerRequest + 'static>(
        &mut self,
    ) -> Result<R::Response, OrsbError> {
        let (req, response_rx) = R::new();
        self.request_tx
            .send(req)
            .await
            .or(Err(OrsbError::BrokerClosed))?;
        response_rx.await.map_err(|_| OrsbError::NoResponse)
    }

    pub async fn subscribe<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        map_broker_response(
            self.make_request::<broker_task::SubscribeRequest<T>>()
                .await,
        )
    }

    pub async fn advertise<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        map_broker_response(
            self.make_request::<broker_task::AdvertiseRequest<T>>()
                .await,
        )
    }

    pub fn subscribe_blocking<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        block_on(self.subscribe::<T>())
    }

    pub fn advertise_blocking<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        block_on(self.advertise::<T>())
    }

    pub async fn claim_receiver<T: Message>(&mut self) -> Result<Receiver<T>, ClaimReceiverError> {
        self.make_request::<broker_task::ClaimReceiverRequest<T>>()
            .await
            .map_err(ClaimReceiverError::Other)
            .and_then(|inner_result| inner_result.map_err(ClaimReceiverError::from))
    }

    pub async fn sender<T: Message>(&mut self) -> Result<Sender<T>, OrsbError> {
        map_broker_response(
            self.make_request::<broker_task::SenderRequestNoWait<T>>()
                .await,
        )
    }

    pub async fn wait_sender<T: Message>(&mut self) -> Result<Sender<T>, OrsbError> {
        map_broker_response(
            self.make_request::<broker_task::SenderRequestWait<T>>()
                .await,
        )
    }

    pub fn claim_receiver_blocking<T: Message>(
        &mut self,
    ) -> Result<Receiver<T>, ClaimReceiverError> {
        block_on(self.claim_receiver::<T>())
    }

    pub fn sender_blocking<T: Message>(&mut self) -> Result<Sender<T>, OrsbError> {
        block_on(self.sender::<T>())
    }

    pub fn wait_sender_blocking<T: Message>(&mut self) -> Result<Sender<T>, OrsbError> {
        block_on(self.wait_sender::<T>())
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[derive(Clone, Debug, PartialEq)]
    struct TestMsg(u8);

    fn init_logger() {
        let _ = env_logger::try_init();
    }

    #[tokio::test]
    async fn test_pub_sub_sync() {
        init_logger();

        let mut orsb = Orsb::start_new();

        tokio::task::spawn_blocking(move || {
            let mut sub = orsb.subscribe_blocking::<TestMsg>().unwrap();
            let mut publ = orsb.advertise_blocking::<TestMsg>().unwrap();

            publ.publish(TestMsg(10)).unwrap();
            publ.publish(TestMsg(20)).unwrap();
            publ.publish(TestMsg(30)).unwrap();

            assert_eq!(sub.recv_blocking(), Ok(TestMsg(10)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(20)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(30)));
            assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_pub_sub_async() {
        init_logger();

        let mut orsb = Orsb::start_new();

        let mut sub = orsb.subscribe::<TestMsg>().await.unwrap();
        let mut publ = orsb.advertise::<TestMsg>().await.unwrap();

        publ.publish(TestMsg(10)).unwrap();
        publ.publish(TestMsg(20)).unwrap();
        publ.publish(TestMsg(30)).unwrap();

        assert_eq!(sub.recv().await, Ok(TestMsg(10)));
        assert_eq!(sub.recv().await, Ok(TestMsg(20)));
        assert_eq!(sub.recv().await, Ok(TestMsg(30)));
        assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
    }

    #[tokio::test]
    async fn test_mpsc_sync() {
        init_logger();

        let mut orsb = Orsb::start_new();

        tokio::task::spawn_blocking(move || {
            let mut recv = orsb.claim_receiver_blocking::<TestMsg>().unwrap();
            assert_eq!(
                orsb.claim_receiver_blocking::<TestMsg>().unwrap_err(),
                ClaimReceiverError::AlreadyClaimed
            );
            let mut send = orsb.sender_blocking::<TestMsg>().unwrap();

            send.send_blocking(TestMsg(10)).unwrap();
            send.send_blocking(TestMsg(20)).unwrap();
            send.send_blocking(TestMsg(30)).unwrap();

            assert_eq!(recv.recv_blocking(), Ok(TestMsg(10)));
            assert_eq!(recv.recv_blocking(), Ok(TestMsg(20)));
            assert_eq!(recv.recv_blocking(), Ok(TestMsg(30)));
            assert_eq!(recv.try_recv(), Err(ReceiverError::Empty));
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_mpsc_async() {
        init_logger();

        let mut orsb = Orsb::start_new();

        let mut recv = orsb.claim_receiver::<TestMsg>().await.unwrap();
        assert_eq!(
            orsb.claim_receiver::<TestMsg>().await.unwrap_err(),
            ClaimReceiverError::AlreadyClaimed
        );
        let mut send = orsb.sender::<TestMsg>().await.unwrap();

        send.send(TestMsg(10)).await.unwrap();
        send.send(TestMsg(20)).await.unwrap();
        send.send(TestMsg(30)).await.unwrap();

        assert_eq!(recv.recv().await, Ok(TestMsg(10)));
        assert_eq!(recv.recv().await, Ok(TestMsg(20)));
        assert_eq!(recv.recv().await, Ok(TestMsg(30)));
        assert_eq!(recv.try_recv(), Err(ReceiverError::Empty));
    }
}
